@@ -10,13 +10,15 @@ import time
 import sys
 from contextlib import asynccontextmanager, suppress, nullcontext, aclosing
 from traceback import format_exception
+from platform import uname
+from functools import partial
 
 import outcome
 from attrs import define,field
 
 from mqttproto import RetainHandling
 from moat.lib.cmd.base import MsgSender, Caller
-from moat.util import CtxObj, P, Root, ValueEvent, timed_ctx, gen_ident, ungroup, NotGiven,Path, PathLongener
+from moat.util import CtxObj, P, Root, ValueEvent, timed_ctx, gen_ident, ungroup, NotGiven,Path, PathLongener,srepr
 from moat.util.compat import CancelledError
 from moat.util.random import al_unique
 
@@ -30,6 +32,7 @@ from .node import Node
 from typing import TYPE_CHECKING,overload
 if TYPE_CHECKING:
     from typing import Any
+    from moat.link.node.codec import CodecNode
 
 class _Requeue(Exception):
     pass
@@ -101,6 +104,9 @@ class BasicCmd:
 
 
 class LinkCommon(CmdCommon):
+    """
+    Common code for links to MQTT only (BaseLink) and a MoaT-Link server (Link).
+    """
     protocol_version:int = -1
     name:str
     server_name:str
@@ -147,8 +153,9 @@ class LinkCommon(CmdCommon):
         self._hello = Hello(me=self.name, me_server=self.is_server, auth_out=auth_out)
         yielded = False
 
-
         async with conn_ as conn:
+            if self.logger.isEnabledFor(logging.INFO):
+                print(f"Connection {self.name} to {srepr(remote)}", file=sys.stderr)
             handler = MsgSender(conn)
             if (res := await self._hello.run(handler)) is False:
                 raise AuthError("Initial handshake failed")
@@ -200,16 +207,10 @@ class _Sender(MsgSender):
     """
     Caller_ = ClientCaller
 
+    _codec_tree:CodecNode|None = None
+
     def __init__(self, link:LinkCommon):
         self._link = link
-        self._allowed = set()
-
-    def enable_path(self, path:Path):
-        """
-        This call protects against forgetting to prefix the root path, or
-        similar bugs, when calling to the back-end directly.
-        """
-        self._allowed.add(path)
 
     @property
     def root(self):
@@ -411,7 +412,6 @@ class _Sender(MsgSender):
 
         The caller is responsible for prefixing the Root path, if applicable.
         """
-        self._pcheck(path, empty_ok=kw.get("subtree",False))
         return self._link.backend.monitor(path, *a, **kw)
 
     def send(self, path:Path, *a, **kw) -> Awaitable[None]:
@@ -423,29 +423,8 @@ class _Sender(MsgSender):
 
         The caller is responsible for prefixing the Root path, if applicable.
         """
-        self._pcheck(path)
         return self._link.backend.send(path, *a, **kw)
 
-    def _pcheck(self,path:Path, empty_ok=False):
-        if len(path) == 0:
-            if empty_ok:
-                return
-            raise ValueError("Empty path?")
-        if isinstance(path[0],Path):
-            if path[0] == Root.get():
-                return
-            if path[0] in self._allowed:
-                return
-            raise ValueError("Prefix not allowed",path[0])
-        root = Root.get()
-        if path[:len(root)] == root:
-            return
-        for p in self._allowed:
-            if path[:len(p)] == p:
-                return
-        raise ValueError("Path not allowed",path)
-
-
     async def i_sync(self):
         """
         This call tries to ensure that the server has processed all
@@ -468,6 +447,29 @@ class _Sender(MsgSender):
         await self.cmd(P("i.sync"), st)
 
 
+    async def _get_tree(self, path, *, task_status, **kw):
+        async with self.d_watch(path, **kw) as w:
+            task_status.started(await w.get_node())
+            await anyio.sleep_forever()
+
+    async def get_codec_tree(self):
+        """
+        Returns a common subtree for codecs.
+        """
+
+        if self._codec_tree is None:
+            self._codec_tree = evt = anyio.Event()
+
+            from moat.link.node.codec import CodecNode
+            self._codec_tree = await self._link.tg.start(partial(self._get_tree, P("codec"), subtree=True, state=None,meta=False, cls=CodecNode))
+            evt.set()
+
+        elif isinstance(self._codec_tree,anyio.abc.Event):
+            await self._codec_tree.wait()
+
+        return self._codec_tree
+
+
 class Link(LinkCommon, CtxObj):
     """
     This class combines the back-end link with a connection to a MoaT-Link server.
@@ -529,6 +531,7 @@ class Link(LinkCommon, CtxObj):
         "Stop me"
         self.tg.cancel_scope.cancel()
 
+
     async def _run_server_link(self, *, task_status=anyio.TASK_STATUS_IGNORED):
         """
         This is the manager task for the server link channel.
@@ -639,7 +642,16 @@ class Link(LinkCommon, CtxObj):
         if isinstance(link, dict):
             link = (link,)
 
+        local_ok = srv.data.get("node","localhost")
+        locals = { "localhost", "127.0.0.1", "::1" }
+
         for remote in link:
+            if local_ok:
+                pass
+            elif "host" not in remote:
+                continue
+            elif remote["host"] in locals:
+                continue
             try:
                 async with timed_ctx(
                     self.cfg.client.init_timeout, self._connect_one(remote, srv.data)
@@ -651,7 +663,7 @@ class Link(LinkCommon, CtxObj):
 
 class BasicLink(LinkCommon, CtxObj):
     """
-    Simple direct link to a server.
+    Simple direct link to a MQTT server.
     """
 
     def __init__(self, cfg, name: str | None, data: dict, **kw):
@@ -691,7 +703,9 @@ class BasicLink(LinkCommon, CtxObj):
 @define(eq=False)
 class Watcher(CtxObj):
     """
-    Helper class for monitoring.
+    Helper class for monitoring (and coalescint the data of) either-or-both
+    * a MQTT subscription to a subtree of our MoaT-Link hierarchy
+    * a MoaT-Link request to enumerate a subtree
     """
     link:Link=field()
     path:Path=field()
@@ -752,13 +766,15 @@ class Watcher(CtxObj):
     async def _ctx(self):
         async with anyio.create_task_group() as tg:
             self._node = self.node_cls()
+            self._current_done = anyio.Event()
             self._tg = tg
             self._qw,self._qr = anyio.create_memory_object_stream(10)
             if self.state is not True:
                 await tg.start(self._updates)
             if self.state is not False:
-                self._current_done = anyio.Event()
                 await tg.start(self._current)
+            else:
+                self._current_done.set()
             yield self
             tg.cancel_scope.cancel()
             await self._qw.aclose()
@@ -805,7 +821,7 @@ class Watcher(CtxObj):
                 return None
             p,d,m = msg
             if self.age is None or m.timestamp+self.age >= time.time():
-                if self._node.set(p,d,m):
+                if self._node.set(p,d,m, force=self._current_done.is_set()):
                     if self.meta:
                         return (p,d,m) if self.subtree else (d,m)
                     else:
@@ -813,7 +829,11 @@ class Watcher(CtxObj):
 
 class Walker:
     """
-    A trimmed-down watcher that retrieves a possibly-partial subtree from the server.
+    A trimmed-down watcher that retrieves a possibly-partial subtree
+    from our MoaT-Link server.
+
+    This differs from `Watcher` by not tracking updates, nor keeping a node
+    tree in memory.
     """
     def __init__(self,mon,meta=False):
         self.mon=mon
